package com.nisum.product.service;

import com.mongodb.MongoClientSettings;
import com.mongodb.client.MongoClient;
import com.mongodb.client.MongoClients;
import com.mongodb.client.MongoCollection;
import com.mongodb.client.MongoDatabase;
import com.mongodb.client.result.InsertOneResult;
import com.nisum.product.dto.FileUploadAttributes;
import com.nisum.product.dto.KafkaPublisherResponse;
import com.nisum.product.dto.ProductOffers;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.io.FileUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.spark.sql.*;
import org.apache.spark.sql.streaming.OutputMode;
import org.apache.spark.sql.streaming.StreamingQuery;
import org.apache.spark.sql.streaming.StreamingQueryException;
import org.apache.spark.sql.streaming.Trigger;
import org.apache.spark.sql.types.*;
import org.apache.spark.util.CollectionAccumulator;
import org.bson.codecs.configuration.CodecRegistries;
import org.bson.codecs.configuration.CodecRegistry;
import org.bson.codecs.pojo.PojoCodecProvider;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.core.io.buffer.DataBufferUtils;
import org.springframework.http.codec.multipart.FilePart;
import org.springframework.stereotype.Service;
import reactor.core.publisher.Flux;

import java.io.File;
import java.io.IOException;
import java.io.Serializable;
import java.nio.charset.StandardCharsets;
import java.text.SimpleDateFormat;
import java.util.*;
import java.util.concurrent.TimeoutException;

@Slf4j
@Service
public class ProductsOfferService implements Serializable {


    @Autowired
    SparkSession sparkSession;

    public Flux<KafkaPublisherResponse> saveAndProcess(Flux<FilePart> filePartFlux) {

        return filePartFlux.flatMap(filePart ->
                filePart.content().map(dataBuffer -> {
                    byte[] bytes = new byte[dataBuffer.capacity()];
                    dataBuffer.read(bytes);
                    DataBufferUtils.release(dataBuffer);

                    return FileUploadAttributes.builder()
                            .fileContent(new String(bytes, StandardCharsets.UTF_8))
                            .fileName(filePart.filename())
                            .build();

                })
                        .map(content -> {
                            return saveAsFile(content);
                        })
                        .map(fileName -> {
                            return readAndPublishDataToKafka(fileName);
                        })
                        .map(kafkaPublisherResponse -> kafkaPublisherResponse));
    }

    private String saveAsFile(FileUploadAttributes fileUploadAttributes) {

        try {
            FileUtils.deleteDirectory(new File("src/main/resources/data/CheckPointLocation"));
        } catch (IOException e) {
            log.error("Error while deleting checkpoint location dir {}", e.getMessage());
        }
        String batchId= new SimpleDateFormat("HHmmssSSS").format(new Date());
        String dir="src/main/resources/data/"+batchId;
        String fileName = dir+ File.separator + fileUploadAttributes.getFileName();
        try {
            FileUtils.forceMkdir(new File(dir));
            FileUtils.writeStringToFile(new File(fileName)
                    , fileUploadAttributes.getFileContent(), StandardCharsets.UTF_8, true);

            return batchId;
        } catch (IOException e) {
            if (e instanceof IOException)
                log.error("Error while creating CSV... {}", e.getMessage());
            return StringUtils.EMPTY;
        }

    }

    private KafkaPublisherResponse readAndPublishDataToKafka(String batchId) {

        if (StringUtils.isNotEmpty(batchId)) {
            String topicName = "test-product2";
            try {
                populateProductsOfferPriceUsingDFStruct(batchId, topicName);
                return KafkaPublisherResponse.builder()
                        .topicName(topicName)
                        .batchId(batchId)
                        .status("File data pushed to Kafka topic").build();
            } catch (TimeoutException | StreamingQueryException e) {
                if (e instanceof TimeoutException)
                    log.error("TimeoutException... {}", e.getMessage());
                if (e instanceof StreamingQueryException)
                    log.error("StreamingQueryException  {}", e.getMessage());
                return KafkaPublisherResponse.builder().topicName(topicName).status(e.getMessage()).build();
            }
        }else
        {
            return KafkaPublisherResponse.builder().build();
        }
    }

    public void populateProductsOfferPriceUsingDF()
    {
        sparkSession.sqlContext().udf()
                .register("calculateOfferPrice",(Double price, Double discountPercentage)->
                        ((100-discountPercentage)*price)/100, DataTypes.DoubleType);
        Dataset<Row> df=sparkSession.read().csv("src/main/resources/data/products_without_header.csv").toDF();
        df.withColumn("discount(%)",functions.lit(10.50))
                .withColumn("OfferPrice",functions.callUDF("calculateOfferPrice",
                        functions.col("_c3").cast(DataTypes.DoubleType)
                        ,functions.col("discount(%)").cast(DataTypes.DoubleType)))
                .write()
                .mode(SaveMode.Overwrite)
                .csv("src/main/resources/result/df/product_offer_price");
    }

    public void populateProductsOfferPriceUsingDFStruct(String batchId,String topicName) throws TimeoutException, StreamingQueryException {


        sparkSession.sqlContext().udf()
                .register("calculateOfferPrice",(Double price, Double discountPercentage)->
                        ((100-discountPercentage)*price)/100, DataTypes.DoubleType);

        StructField[] structFields= new StructField[4];
        structFields[0]= new StructField("id",DataTypes.StringType,true, Metadata.empty());
        structFields[1]= new StructField("name",DataTypes.StringType,true,Metadata.empty());
        structFields[2]= new StructField("description",DataTypes.StringType,true,Metadata.empty());
        structFields[3]= new StructField("price",DataTypes.DoubleType,true,Metadata.empty());

        StructType structType= new StructType(structFields);

        Dataset<Row> df=sparkSession.readStream()
                .format("csv")
                .option("header",true)
                .schema(structType)
                .load("src/main/resources/data/"+batchId);

        Dataset<Row> rowTransformDf=df.withColumn("discountPercentage",functions.lit(10.50))
                .withColumn("offerPrice",functions.callUDF("calculateOfferPrice",
                        functions.col("price").cast(DataTypes.DoubleType)
                        ,functions.col("discountPercentage").cast(DataTypes.DoubleType)))
                .withColumn("uniqueId", functions.lit(UUID.randomUUID().toString()))
                .withColumn("batchId",functions.lit(batchId));


        rowTransformDf.selectExpr("CAST(BatchId AS STRING) AS key", "to_json(struct(*)) AS value")
                .writeStream()
                .format("kafka")
                .trigger(Trigger.Once())
                .option("kafka.bootstrap.servers", "HYD-LAP-00484.corp.nisum.com:9092")
                .option("topic", topicName)
                .option("checkpointLocation", "src/main/resources/data/CheckPointLocation")
                .start()
                .awaitTermination(5000);
    }



    public Flux<ProductOffers> getProductOffers(String batchId) throws TimeoutException, StreamingQueryException {


        Dataset<Row> df = sparkSession
                .readStream()
                .format("kafka")
                .option("kafka.bootstrap.servers", "HYD-LAP-00484.corp.nisum.com:9092")
                .option("subscribe", "test-product2")
                .option("startingOffsets", "earliest")
                                .option("startingOffsets", "earliest")
                                .option("failOnDataLoss", "false").
                    option("kafka.con.group.id", "localStream")
                .load();

        StructType productOffers = Encoders.bean(ProductOffers.class).schema();

        Dataset<ProductOffers> offersDataset=df.selectExpr("CAST(value as String) as message")
                .select(functions.from_json(functions.col("message"),productOffers ).as("data"))
                .filter(functions.col("data.batchId").equalTo(batchId))
                .filter(functions.col("data.name").contains("Apple"))
                .filter(functions.col("data.offerPrice").geq(20000))
                .select("data.*")
                .coalesce(1)
                .as(Encoders.bean(ProductOffers.class));



        StreamingQuery streamingQuery=offersDataset
                .writeStream()
                .trigger(Trigger.Once())
                .format("csv")
                .option("checkpointLocation", "src/main/resources/data/checkpointLocation/StreamingJob")
                .option("path","src/main/resources/data/"+batchId)
                .outputMode(OutputMode.Append())
                .start();


        CollectionAccumulator<ProductOffers> offerProductList=sparkSession.sparkContext().collectionAccumulator("offerProductList");




        StreamingQuery query=offersDataset
                .writeStream()
                .trigger(Trigger.Once())
                .foreach(new ForeachWriter<ProductOffers>() {

                    @Override
                    public boolean open(long partitionId, long epochId) {
                        return true;
                    }

                    @Override
                    public void process(ProductOffers value) {

                        offerProductList.add(value);

                    }

                    @Override
                    public void close(Throwable errorOrNull) {

                    }
                }).start();


        query.awaitTermination();


        streamingQuery.awaitTermination();

        try {
            log.info("Offer Product List : {}",offerProductList);
            log.info("Awaiting main thread ......");
            Thread.sleep(2000);
            log.info("Resuming main thread ......");
        } catch (InterruptedException e) {
            e.printStackTrace();
        }


        return Flux.fromIterable(offerProductList.value());

    }






}
